(* ************************************************************************** *)
(*           Read/Write form status information from/to .INI file             *)
(*                  Copyright (C) 1998-2004 Fumio Kawamata                    *)
(*                                                                            *)
(*  You can get Version Information of the program by using this unit.        *)
(*                                                                            *)
(*  This program is free software; you can redistribute it and/or modify      *)
(*  it under the terms of the GNU General Public License as published by      *)
(*  the Free Software Foundation; version 2 of the License.                   *)
(*                                                                            *)
(*  This program is distributed in the hope that it will be useful, but       *)
(*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      *)
(*  ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General         *)
(*  Public License for more details.  You must be have it.                    *)
(* ************************************************************************** *)

unit FormStatus;

interface

uses Forms;

procedure ReadFormStatusFromIniFile(Form: TForm; IniFileFullPath: String);
procedure WriteFormStatusToIniFile(Form: TForm; IniFileFullPath: String);
function ReadIntegerFromIniFile(Section,Ident: String; Default: Longint; IniFileFullPath: String): Longint;
procedure WriteIntegerToIniFile(Section,Ident: String; Value: Longint; IniFileFullPath: String);
function ReadStringFromIniFile(Section,Ident: String; Default: String; IniFileFullPath: String): String;
procedure WriteStringToIniFile(Section,Ident: String; Value: String; IniFileFullPath: String);
function ReadBoolFromIniFile(Section,Ident: String; Default: Boolean; IniFileFullPath: String): Boolean;
procedure WriteBoolToIniFile(Section,Ident: String; Value: Boolean; IniFileFullPath: String);



implementation

uses IniFiles;

procedure ReadFormStatusFromIniFile(Form: TForm; IniFileFullPath: String);
var
  AppIni : TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  Form.Top:=AppIni.ReadInteger(Form.Name,'Top', Form.Top);
  Form.Left:=AppIni.ReadInteger(Form.Name,'Left', Form.Left);
  Form.Width:=AppIni.ReadInteger(Form.Name,'Width', Form.Width);
  Form.Height:=AppIni.ReadInteger(Form.Name,'Height', Form.Height);
  Form.WindowState:=TWindowState(AppIni.ReadInteger(Form.Name,'WindowState', 0));
  AppIni.Free;
end;

procedure WriteFormStatusToIniFile(Form: TForm; IniFileFullPath: String);
var
  AppIni : TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  AppIni.WriteInteger(Form.Name,'WindowState', Integer(Form.Windowstate));
  AppIni.WriteInteger(Form.Name,'Top', Form.Top);
  AppIni.WriteInteger(Form.Name,'Left', Form.Left);
  AppIni.WriteInteger(Form.Name,'Width', Form.Width);
  AppIni.WriteInteger(Form.Name,'Height', Form.Height);
  AppIni.UpdateFile;
  AppIni.Free;
end;

function ReadIntegerFromIniFile(Section,Ident: String; Default: Longint; IniFileFullPath: String): Longint;
var
  AppIni: TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  Result:=AppIni.ReadInteger(Section, Ident, Default);
  AppIni.Free;
end;

procedure WriteIntegerToIniFile(Section,Ident: String; Value: Longint; IniFileFullPath: String);
var
  AppIni : TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  AppIni.WriteInteger(Section, Ident, Value);
  AppIni.UpdateFile;
  AppIni.Free;
end;


function ReadStringFromIniFile(Section,Ident: String; Default: String; IniFileFullPath: String): String;
var
  AppIni: TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  Result:=AppIni.ReadString(Section, Ident, Default);
  AppIni.Free;
end;

procedure WriteStringToIniFile(Section,Ident: String; Value: String; IniFileFullPath: String);
var
  AppIni : TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  AppIni.WriteString(Section, Ident, Value);
  AppIni.UpdateFile;
  AppIni.Free;
end;

function ReadBoolFromIniFile(Section,Ident: String; Default: Boolean; IniFileFullPath: String): Boolean;
var
  AppIni: TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  Result:=AppIni.ReadBool(Section, Ident, Default);
  AppIni.Free;
end;

procedure WriteBoolToIniFile(Section,Ident: String; Value: Boolean; IniFileFullPath: String);
var
  AppIni : TMemIniFile;
begin
  AppIni:=TMemIniFile.Create(IniFileFullPath);
  AppIni.WriteBool(Section, Ident, Value);
  AppIni.UpdateFile;
  AppIni.Free;
end;


end.
