(*
 *
 *              A Mail Filtering and Control Utility   "MailUtl"
 *                     Copyright (C) 2002-2005  Fumio KAWAMATA
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.
 *
 *)
unit MailUtl2_Main;

interface

{$DEFINE Japanese}
(* Remove above compiler directive if you do not use Japanese *)

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, ExtCtrls, Menus, Grids, MailUtl2_DM, IdBaseComponent,
  IdComponent, IdTCPConnection, IdTCPClient, IdExplicitTLSClientServerBase,
  IdMessageClient, IdMessage, IdPOP3, IdReplyPOP3, IdException,IdStrings,
  IdGlobalProtocols, IdGlobal, IdSys;
  (* Indy.Sockets Version 10.0.52 Copyright (c) 1993 - 2004,
   * Chad Z. Hower (Kudzu) and the Indy Pit Crew
   *)

type
  TMyRunningStatus =(rsIdle, rsChecking, rsOnLineBrowsing, rsOffLineBrowsing, rsSorting, rsDeleting);

  TFormMain = class(TForm)
    MainMenu: TMainMenu;
    StatusBar: TStatusBar;
    Splitter: TSplitter;
    PanelActionReason: TPanel;
    EditActionReason: TEdit;
    TabControlMain: TTabControl;
    Memo: TMemo;
    MainMenuFile: TMenuItem;
    MainMenuHelp: TMenuItem;
    MainMenuHelpVersion: TMenuItem;
    MainMenuFileExit: TMenuItem;
    IdPOP3: TIdPOP3;
    MainMenuOption: TMenuItem;
    MainMenuOptionSettings: TMenuItem;
    MainMenuOptionFont: TMenuItem;
    FontDialog: TFontDialog;
    MainMenuOptionIniFileName: TMenuItem;
    MainMenuOptionSeparator: TMenuItem;
    MainMenuFileCheck: TMenuItem;
    PanelStatusBar: TPanel;
    ImageBulbeOn: TImage;
    ImageBulbeOff: TImage;
    LabelConnectionStatus: TLabel;
    ProgressBar: TProgressBar;
    MainMenuFileDelete: TMenuItem;
    MainMenuFileDisconnect: TMenuItem;
    MainMenuEdit: TMenuItem;
    MainMenuEditRuleFile: TMenuItem;
    MainMenuSearch: TMenuItem;
    MainMenuSearchTop: TMenuItem;
    MainManuSearchLast: TMenuItem;
    MainMenuFileAbort: TMenuItem;
    TimerNoop: TTimer;
    MainMenuFileDeleteTest: TMenuItem;
    TimerClearStatus: TTimer;
    MainMenuSort: TMenuItem;
    MainMenuSortByAction: TMenuItem;
    MainMenuSortByNumber: TMenuItem;
    PanelStringGrid: TPanel;
    StringGrid: TStringGrid;
    procedure FormResize(Sender: TObject);
    procedure MainMenuHelpVersionClick(Sender: TObject);
    procedure MainMenuFileExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MainMenuOptionSettingsClick(Sender: TObject);
    procedure MainMenuOptionFontClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure StringGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure StringGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure TabControlMainChange(Sender: TObject);
    procedure TimerNoopTimer(Sender: TObject);
    procedure MainMenuFileAbortClick(Sender: TObject);
    procedure MainMenuFileDisconnectClick(Sender: TObject);
    procedure TimerClearStatusTimer(Sender: TObject);
    procedure MainMenuFileDeleteClick(Sender: TObject);
    procedure MainMenuFileDeleteTestClick(Sender: TObject);
    procedure MainMenuEditRuleFileClick(Sender: TObject);
    procedure MainMenuSearchTopClick(Sender: TObject);
    procedure MainManuSearchLastClick(Sender: TObject);
    procedure MainMenuSortByActionClick(Sender: TObject);
    procedure MainMenuSortByNumberClick(Sender: TObject);
    procedure MainMenuFileCheckClick(Sender: TObject);
    procedure CheckMailBox;
    procedure DeleteMails(ShamMode: Boolean);
  private
    procedure DisplayCaptionOfStringGrid;
    procedure DisplayConnectionStatus(Status: Boolean);
    procedure EnableOrDisableControls(RunningStatus: TMyRunningStatus);
    procedure SelectRow(Row: Integer);
    procedure ReadStringGridStatusFromIniFile(IniFileName: string);
    procedure TimeredDisplayInStatusBar(StatusString: String; DisplaySeconds: Integer);
    procedure WriteStringGridStatusToIniFile(IniFileName: string);
  public
  end;

  TMyAction = (actNop, actKeep, actDel);

  TMyEvaluated=record
    Action: TMyAction;
    Reason: String;
  end;


var
  FormMain: TFormMain;

implementation

{$R *.dfm}

uses
  VerInfo, FontInfo, FormStatus,
  {$IFDEF Japanese}
  PzConv,
  (* PzConv.pas Copyright (C) 1998 Pz factory All rights Reserved. *)
  (* PzConv.pas is a text converter. *)
  (* PzConv.pas includes Japanese. *)
  {$ENDIF Japanese}
  BRegExp,
  (* BregExp.pas  Copyright (C) Osamu Takeuchi All rights Reserved. *)
  (* BregExp.pas includes Japanese. *)
  (* BregExp.pas requires BregExp.dll, Copyright (C) Tatsuo Baba *)
  IniFiles, ShellAPI, MailUtl2_Global, MailUtl2_Settings;

const
  CRLF = chr($0D)+chr($0A);
  RegExpOfIPAddress='m/[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/';

var
  {$IFDEF Japanese}
  Jap: TPzConv;
  {$ENDIF Japanese}
  Msg: TMessage = (Msg: CM_PARENTFONTCHANGED);
  (* TMessage requires the "Messages" unit. *)
  (* CM_PARENTFONTCHANGED requires the "Controls" unit. *)
  Field1List: TStringList;
  Field2List: TStringList;
  Field3List: TStringList;
  RunningStatus: TMyRunningStatus;
  ForcedTermination: Boolean;
  Connected: Boolean;
  LogFileName: String;



(*
 * Initialize values
 *)
procedure Initialize;
begin
  AutoRun:=False;
  Connected:=False;
  ExeDir:=ExtractFileDir(ParamStr(0));
  IniFileFullPath:=ChangeFileExt(ParamStr(0), '.ini');
  LogDir:=SysUtils.GetEnvironmentVariable('TEMP');
  MatchString:='';
  MaxLines:=20;
  POP3_Host:='';
  POP3_Port:=110;
  POP3_UserName:='';
  POP3_Password:=SysUtils.GetEnvironmentVariable('POP3_USERPASS');
  RuleFileName:='';
  RunningStatus:=rsIdle;
  UseMailUtlRules:=False;
  //LogFileBase:='MailUtl-'+FormatDateTime('yyyymmdd-hhnnss', Now);
end;



(* *****************************************************************************
 * Get command line options
 * *****************************************************************************
 *)
procedure GetCommandLineOptions;
var
  I: Integer;
  S: String;
begin
  if ParamCount=0 then Exit;
  for I:=1 to ParamCount do
  begin
    S:=ParamStr(I);
    if Pos('-I', S)=1 then
    begin
      Delete(S, 1, 2);
      IniFileFullPath:=S;
    end
    else
    if Pos('-AUTO', S)=1 then
    begin
      Delete(S, 1, 2);
      AutoRun:=True;
    end
  end;
end;



(* *****************************************************************************
 * SetShortMonthNames
 *   ShortMonthNames[] is a built in value.
 * *****************************************************************************
 *)
procedure SetShortMonthNames;
const
  ShortMonthNamesArray: array[1..12] of string = ('Jan', 'Feb', 'Mar', 'Apr',
    'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
var
  I: Integer;
begin
  for I:=1 to 12 do
    ShortMonthNames[I]:=ShortMonthNamesArray[I];
end;




//*****************************************************************************
//* Logging
//*   Logging message into the log file
//*****************************************************************************
function Logging(Msg: string): Boolean;
var
  LogFile:     Text;
  LogDateTime: TDateTime;
  Errored:     Boolean;
begin
  LogDateTime:=Now;
  Logging:=True;
  Assign(LogFile, LogFileName);
  Errored:=False;
  if FileExists(LogFileName) then
    Append(LogFile)
  else
  begin
    try
      Rewrite(LogFile);
    except
      Errored:=True;
    end;
    if Errored=False then
      Writeln(LogFile, '# MailUtl log file');
  end;

  if Errored then
    Logging:=False
  else
  begin
    if Msg='' then
      Writeln(LogFile)
    else
      Writeln(LogFile, FormatDateTime('mmm/dd/yyyy hh:nn:ss ', LogDateTime)+Msg);
    CloseFile(LogFile);
  end;
end;



//*****************************************************************************
//* GetFieldValue
//*   finds designated field as "Field" constant,
//*   returns field value.
//*****************************************************************************
function GetFieldValue(S: string; Field: string): string;
var
  Index: Integer;
  Count: Integer;
begin
  Result:='';
  if (pos(UpperCase(Field), UpperCase(S) )=1) and (Length(S)>Length(Field)) then
  begin
    Index:=Length(Field)+1;
    Count:=Length(S)-Length(Field);
    Result:=Trim(Copy(S, Index, Count))
  end;
end;




//*****************************************************************************
//* GetFieldValueInList
//*   finds designated field as "Field" constant in the stringlist,
//*   returns field value.
//*   Note: If two or more fields are found, this function returns
//*         all mutched fields.  (Ex: To:, Cc:, From: or Received: field)
//*****************************************************************************
function GetFieldValueInList(Headers: TStringList; Field: string): string;
var
  FieldValue: string;
  Finish:     Boolean;
  I:          Integer;
  S:          string;
begin
  Result:='';
  Finish:=False;
  I:=0;
  while (I<Headers.Count)and(Finish=False) do
  begin
    S:=Headers[I];
    FieldValue:=GetFieldValue(S, Field);
    if FieldValue<>'' then
    begin
      if Result<>'' then
        Result:=Result+' , ';
      Result:=Result+FieldValue;
      Finish:=True;
      S:=Uppercase(Field);
      if (S='FROM:')or(S='TO:')or(S='CC:')or(S='REPLY-TO:')or(S='RECEIVED:') then
        Finish:=False;
    end;
    inc(I);
  end;
end;




//*****************************************************************************
//* ArrangeHeaders
//*   converts headers to Japanese,
//*   combine header lines which consist of several rows (ex: "Received:"
//*   header) into one line.
//*****************************************************************************
procedure ArrangeHeaders(AMsg: TIdmessage);
var
  Count:           Integer;
  FirstChar:       char;
  I:               Integer;
  LastField:       string;
  NumberOfHeaders: Integer;
  UpperStr:        string;
  S:               String;
  TmpStr:          string;
begin
  NumberOfHeaders:=AMsg.Headers.Count;
  LastField:='';
  I:=0; Count:=0;
  repeat
    TmpStr:=AMsg.Headers[I];
    {$IFDEF Japanese}
    TmpStr:=Jap.MIMEHeaderDecode(TmpStr);
    TmpStr:=Jap.JisToSJis(TmpStr);
    {$ENDIF Japanese}
    UpperStr:=UpperCase(TmpStr);
    FirstChar:=TmpStr[1];
    if (FirstChar=' ')or(FirstChar=char($09)) then
    begin
      if LastField='SUBJECT:' then
        AMsg.Headers[I-1]:=AMsg.Headers[I-1]+Trim(TmpStr)
      else
        AMsg.Headers[I-1]:=AMsg.Headers[I-1]+' '+Trim(TmpStr);
      AMsg.Headers.Delete(I);
    end
    else
    begin
      AMsg.Headers[I]:=TmpStr;
      if pos('SUBJECT:', UpperStr)=1 then
        LastField:='SUBJECT:'
      else
        LastField:='';
      inc(I);
    end;
    inc(Count);
  until Count>=NumberOfHeaders;
  S:=GetFieldValueInList(AMsg.Headers, 'Date:');
  try
    AMsg.Date:=GMTToLocalDateTime(S);
  except
    Logging('*** Error *** Date: '+S);
    AMsg.Date:=0.0;
  end;
  S:=GetFieldValueInList(AMsg.Headers, 'From:');
  brx.Subst('s/[' + #13 + #10 + ']//g', S);
  Logging('From: '+S);
  try
    AMsg.From.Text:=S;
  except
    Logging ('*** Error in From: ***');
  end;
  AMsg.Subject:=GetFieldValueInList(AMsg.Headers, 'Subject:');
  Logging('Subject: '+AMsg.Subject);
  AMsg.ReplyTo.EMailAddresses:=GetFieldValueInList(AMsg.Headers, 'Reply-To:');
  AMsg.Recipients.EMailAddresses:=GetFieldValueInList(AMsg.Headers, 'To:');
end;




//*****************************************************************************
//* SplitBeforeDelimiter
//*   Split string by delimiter
//*****************************************************************************
function SplitBeforeDelimiter(var S: string; Delimiter: Char): string;
var
  TabPos: Integer;
begin
  TabPos:=Pos(Delimiter, S);
  if TabPos=0 then
  begin
    Result:=Trim(S);
    S:='';
  end
  else
  if TabPos=1 then
  begin
    Result:='';
    Delete(S, 1, 1);
  end
  else
  begin
    Result:=Trim(Copy(S, 1, TabPos-1));
    Delete(S, 1, TabPos);
  end;
end;




//*****************************************************************************
//* ReadRuleFromFile
//*   Read rule file from the file
//*****************************************************************************
function ReadRuleFromFile(RuleFileName: string): Boolean;
var
  F:         Text;
  S:         string;
  SearchStr: string;
  AFieldStr: string;
  FieldsStr: string;
  ActionStr: string;
begin
  ReadRuleFromFile:=True;
  try
    Assign(F, RuleFileName);
    Reset(F);
  except
    ReadRuleFromFile:=False;
    Exit;
  end;
  while not(eof(F)) do
  begin
    Readln(F, S);
    S:=Trim(S);
    if (Pos('#', S)<>1)and(S<>'') then
    begin
      SearchStr:=SplitBeforeDelimiter(S, Chr($09));
      if SearchStr='' then
      begin
        CloseFile(F);
        Exit;
      end;

      FieldsStr:=SplitBeforeDelimiter(S, Chr($09));
      if FieldsStr='' then
      begin
        FieldsStr:='FROM:'; //* Default value
      end;

      ActionStr:=SplitBeforeDelimiter(S, Chr($09));
      if ActionStr='' then
      begin
        ActionStr:='DEL'; //* Default value
      end;

      while FieldsStr<>'' do
      begin
        AFieldStr:=SplitBeforeDelimiter(FieldsStr, ',');
        if (SearchStr<>'')and(AFieldStr<>'')and(ActionStr<>'') then
        begin
          Field1List.Add(SearchStr);
          Field2List.Add(AFieldStr);
          Field3List.Add(ActionStr);
        end;
      end;
    end;
  end;
  CloseFile(F);
end;



//*****************************************************************************
//* CheckField
//*****************************************************************************
function CheckField(Headers: TStringList; Field: string; S: string): Boolean;
var
  FieldValue:    string;
  S2:            string;
  NegativeLogic: Boolean;
  RegularExpression: Boolean;
begin
  RegularExpression:=False;
  NegativeLogic:=False;
  Result:=False;
  S2:=S;
  if pos('!', S2)=1 then
  begin
    NegativeLogic:=True;
    Delete(S2, 1,1);
    S2:=Trim(S2);
  end;
  if pos('\', S2)=1 then
  begin
    Delete(S2, 1, 1);
    RegularExpression:=True;
  end;

  if S2<>'' then
  begin
    FieldValue:=GetFieldValueInList(Headers, Field);
    if RegularExpression=True then
    begin
      Result:=brx.Match(S2, FieldValue);
    end
    else begin
      if pos(UpperCase(S2), UpperCase(FieldValue))>0 then
        Result:=True;
    end;
    if NegativeLogic=True then
    begin
      Result:=not(Result);
    end;
  end;
end;




//*****************************************************************************
//* CheckRule
//*****************************************************************************
function CheckRule(RxMsg: TIdMessage; MsgSize: Integer): TMyEvaluated;
var
  Finish:         Boolean;
  RuleIndex:      Integer;
  Matched:        Boolean;
  MatchedReason:  string;
  MisDateTime:    TDateTime;
  AllowedMisDate: Integer;
  S:              string;
  Size:           Integer;
begin
  Result.Action:=actNop;
  Result.Reason:='';
  RuleIndex:=0;
  Finish:=False;
  while (RuleIndex<Field1List.Count)and(Finish=False) do
  begin
    Matched:=False;

    //* Message size
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='MSGSIZE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        Size:=StrToIntDef(S, 0);
        if MsgSize>Size then
        begin
          Matched:=True;
          MatchedReason:='Message size ('+IntToStr(MsgSize)+')>'+IntToStr(Size);
        end;
      end;
    end;

    //* Future Date
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='FUTUREDATE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        AllowedMisDate:=StrToIntDef(S, 0);
        MisDateTime:=RxMsg.Date-Now;
        if MisDateTime>AllowedMisDate then
        begin
          Matched:=True;
          MatchedReason:='Misdated ('+DateTimeToStr(RxMsg.Date)+' is future date over '+IntToStr(AllowedMisDate)+' days)';
        end;
      end;
    end;


    //* Past Date
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='PASTDATE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        AllowedMisDate:=StrToIntDef(S, 0);
        MisDateTime:=Now-RxMsg.Date;
        if MisDateTime>AllowedMisDate then
        begin
          Matched:=True;
          MatchedReason:='Misdated ('+DateTimeToStr(RxMsg.Date)+' is past date over '+IntToStr(AllowedMisDate)+' days)';
        end;
      end;
    end;


    if CheckField(RxMsg.Headers, Field2List[RuleIndex], Field1List[RuleIndex])=True then
    begin
      Matched:=True;
      MatchedReason:='"'+Field1List[RuleIndex]+'" in "'+Field2List[RuleIndex]+'"';
    end;

    S:=UpperCase(Field3List[RuleIndex]);

    if Matched=True then
    begin
      if Result.Reason<>'' then
      begin
        Result.Reason:=Result.Reason+' and ';
      end;
      Result.Reason:=Result.Reason+MatchedReason;
      if S<>'CONTINUE' then
      begin
        if S='KEEP' then
        begin
          Result.Action:=actKeep
        end
        else
        if S='DEL' then
        begin
          Result.Action:=actDel
        end
        else
        begin
          Result.Action:=actNop;
        end;
        Finish:=True;
      end;
    end
    else
    begin //* UnMatched
      if S='CONTINUE' then
      begin //* Skip all 'CONTINUE' actions.
        while (RuleIndex<(Field1List.Count-1))and(Uppercase(Field3List[RuleIndex])='CONTINUE') do
        begin
          Inc(RuleIndex);
        end;
      end;
      Result.Reason:='';
    end;

    if Finish=False then
    begin
      if RuleIndex<Field1List.Count then
      begin
        Inc(RuleIndex);
      end;
    end;
  end;
end;




//*****************************************************************************
//* EncodeForHTML
//*****************************************************************************
function EncodeForHTML(S: string): string;
var
  I:   Integer;
  WS:  WideString;
  Tmp: string;
begin
  WS:=S;
  Result:='';
  if WS='' then
    Exit;
  for I:=1 to Length(WS) do
  begin
    Tmp:=copy(WS, I, 1);
    if Tmp='<' then
      Tmp:='&lt;'
    else
    if Tmp='>' then
      Tmp:='&gt;'
    else
    if Tmp='&' then
      Tmp:='&amp;'
    else
    if Tmp='"' then
      Tmp:='&quot;'
    ;
    Result:=Result+Tmp;
  end;
end;



(* *****************************************************************************
 * ReadParametersFromIniFile
 *   Read parameters from the .INI file
 * *****************************************************************************
 *)
function ReadParametersFromIniFile(IniFileName: string): boolean;
var
  IniFile: TIniFile;
begin
  ReadParametersFromIniFile:=True;
  IniFile := TIniFile.Create(IniFileName);
  try
    LogDir:=IniFile.ReadString('General', 'LogDir', LogDir);
    MatchString:=IniFile.ReadString('General', 'MatchString', MatchString);
    MaxLines:=IniFile.ReadInteger('General', 'MaxLines', MaxLines);
    POP3_Host:=IniFile.ReadString('POP3', 'Host', POP3_Host);
    POP3_Port:=IniFile.ReadInteger('POP3', 'Port', POP3_Port);
    POP3_UserName:=IniFile.ReadString('POP3', 'UserID', POP3_UserName);
    RuleFileName:=IniFile.ReadString('General', 'RuleFile', RuleFileName);
    UseMailUtlRules:=IniFile.ReadBool('General', 'UseMailUtlRules', UseMailUtlRules);
  except
    ReadParametersFromIniFile:=False;
  end;
  IniFile.Free;
end;



(* *****************************************************************************
 * WriteParametersToIniFile
 *   Write parameters to the .INI file
 * *****************************************************************************
 *)
function WriteParametersToIniFile(IniFileName: string): boolean;
var
  IniFile: TIniFile;
begin
  WriteParametersToIniFile:=True;
  IniFile := TIniFile.Create(IniFileName);
  try
    IniFile.WriteString('General', 'LogDir', LogDir);
    IniFile.WriteString('General', 'MatchString', MatchString);
    IniFile.WriteInteger('General', 'MaxLines', MaxLines);
    IniFile.WriteString('POP3', 'Host', POP3_Host);
    IniFile.WriteInteger('POP3', 'Port', POP3_Port);
    IniFile.WriteString('POP3', 'UserID', POP3_UserName);
    IniFile.WriteString('General', 'RuleFile', RuleFileName);
    IniFile.WriteBool('General', 'UseMailUtlRules', UseMailUtlRules);
  except
    WriteParametersToIniFile:=False;
  end;
  IniFile.Free;
end;



procedure TFormMain.TimeredDisplayInStatusBar(StatusString: String; DisplaySeconds: Integer);
begin
  TimerClearStatus.Enabled:=False;
  TimerClearStatus.Interval:=DisplaySeconds*1000;
  StatusBar.SimpleText:=StatusString;
  TimerClearStatus.Enabled:=True;
end;



(*
 * Right-justify the Size column and Left-justify the other columns.
 *)
procedure TFormMain.StringGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  Action: Integer;
  SG: TStringGrid;
  X: Integer;
begin
  SG:=TStringGrid(Sender);
  if ARow=0 then
  begin (* Title row *)
    if (ACol=4)or(ACol=9) then
    begin (* Right justify *)
      SetTextAlign(SG.Canvas.Handle, TA_RIGHT or TA_TOP);
      X:=Rect.Right-2;
    end
    else
    begin (* Left justify *)
      SetTextAlign(SG.Canvas.Handle, TA_LEFT or TA_TOP);
      X:=Rect.Left+2;
    end;
    ExtTextOut(SG.Canvas.Handle, X, Rect.Top+2, ETO_CLIPPED or ETO_OPAQUE,
    @Rect, PChar(SG.Cells[ACol, ARow]), Length(SG.Cells[ACol, ARow]), nil);
  end
  else
  begin (* Not title row *)
    Action:=StrToIntDef(SG.Cells[0, ARow],ord(actNop));

    (* Coloring by Action value *)
    if Action=ord(actDel) then
    begin
      SG.Canvas.Font.Color:=clRed;
      SG.Canvas.FillRect(Rect);
    end;

    if (ACol=3)and(Action=ord(actDel)) then
    begin (* Centering *)
      X:=(Rect.Left+Rect.Right) div 2;
      X:=X-8;
      DM.ImageList.Draw(SG.Canvas, X, Rect.Top, 47);
    end
  end;

  if (ACol=4)or(ACol=9) then
  begin (* Right justify *)
    SetTextAlign(SG.Canvas.Handle, TA_RIGHT or TA_TOP);
    X:=Rect.Right-2;
  end
  else
  begin (* Left justify *)
    SetTextAlign(SG.Canvas.Handle, TA_LEFT or TA_TOP);
    X:=Rect.Left+2;
  end;
  if ACol<>3 then
  begin
    ExtTextOut(SG.Canvas.Handle, X, Rect.Top+2, ETO_CLIPPED or ETO_OPAQUE,
      @Rect, PChar(SG.Cells[ACol, ARow]), Length(SG.Cells[ACol, ARow]), nil);
  end;

end;



(*
 * ReadStringGridStatusFromIniFile
 *   Read string grid parameters from the .INI file
 *)
procedure TFormMain.ReadStringGridStatusFromIniFile(IniFileName: string);
var
  I: Integer;
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(IniFileName);
  For I:=0 to StringGrid.ColCount-1 do
  begin
    StringGrid.ColWidths[I]:=IniFile.ReadInteger('StringGrid', 'Col'+IntToStr(I+1), 100);
  end;
  PanelStringGrid.Height:=IniFile.ReadInteger('PanelStringGrid', 'Height', 200);
  IniFile.Free;
end;



(*
 * WriteStringGridStatusToIniFile
 *   Write string grid parameters to the .INI file
 *)
procedure TFormMain.WriteStringGridStatusToIniFile(IniFileName: string);
var
  I: Integer;
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(IniFileName);
  //IniFile.WriteInteger('StringGrid', 'ColCount', StringGrid.ColCount);
  For I:=0 to StringGrid.ColCount-1 do
  begin
    IniFile.WriteInteger('StringGrid', 'Col'+IntToStr(I+1), StringGrid.ColWidths[I]);
  end;
  IniFile.WriteInteger('PanelStringGrid', 'Height', PanelStringGrid.Height);
  IniFile.Free;
end;



procedure TFormMain.DisplayCaptionOfStringGrid;
begin
  StringGrid.Cells[4,0]:='No.';
  StringGrid.Cells[5,0]:='Action';
  StringGrid.Cells[6,0]:='Subject';
  StringGrid.Cells[7,0]:='From address';
  StringGrid.Cells[8,0]:='Date';
  StringGrid.Cells[9,0]:='Size';
end;



procedure TFormMain.DisplayConnectionStatus(Status: Boolean);
begin
  if Status=True then
  begin
    ImageBulbeOn.Visible:=True;
    ImageBulbeOff.Visible:=False;
    LabelConnectionStatus.Caption:='Connected';
  end
  else begin
    ImageBulbeOn.Visible:=False;
    ImageBulbeOff.Visible:=True;
    LabelConnectionStatus.Caption:='Disconnected';
  end;
end;



procedure TFormMain.FormResize(Sender: TObject);
begin
  EditActionReason.Width:=PanelActionReason.Width-10;
  ProgressBar.Left:=StatusBar.Width-ProgressBar.Width-12;
  PanelStatusBar.Left:=StatusBar.Width-(PanelStatusBar.Width+ProgressBar.Width)-16;
end;



procedure TFormMain.MainMenuHelpVersionClick(Sender: TObject);
var
  S: String;
begin
  S:='Portions of this software are Copyright (c) 1993 - 2004, Chad Z. Hower (Kudzu) and the Indy Pit Crew - http://www.nevrona.com/Indy/'+CRLF;
  S:=S+'PzConv.pas Copyright (C) 1998 Pz factory'+CRLF;
  S:=S+'BREGEXP.DLL Copyright (C) Tatsuo Baba'+CRLF;
  S:=S+'BregExp.pas Copyright (C) Osamu Takeuchi';
  VerInfo.DisplayVersionInfo(S);
end;



procedure TFormMain.MainMenuFileExitClick(Sender: TObject);
begin
  if Connected=True then
  begin
    IdPOP3.Disconnect;
  end;
  Close;
end;



procedure TFormMain.FormCreate(Sender: TObject);
var
  AppIni: TIniFile;
  S: String;
begin
  TimerNoop.Enabled:=False;
  PanelStatusBar.Color:=clBtnFace;
  Initialize;
  SetShortMonthNames;
  GetCommandLineOptions;

  if FileExists(IniFileFullPath)=False then
  begin
    AppIni:=TIniFile.Create(IniFileFullPath);
    AppIni.UpdateFile;
    AppIni.Free;
    MessageDlg('The program created "'+IniFileFullPath+'" file.',mtInformation,[mbOK],0);
    FormMain.Position:=poScreenCenter;
  end;

  ReadParametersFromIniFile(IniFileFullPath);
  FormStatus.ReadFormStatusFromIniFile(FormMain, IniFileFullPath);
  ReadStringGridStatusFromIniFile(IniFileFullPath);
  StringGrid.ColWidths[0]:=0;
  StringGrid.ColWidths[1]:=0;
  StringGrid.ColWidths[2]:=0;
  FontInfo.ReadFontInfoFromIniFile(FormMain, IniFileFullPath);
  FormMain.Caption:=FormMain.Caption+'  '+GetVersionInfo('FileVersion');
  S:=FormMain.MainMenuOptionIniFileName.Caption;
  FormMain.MainMenuOptionIniFileName.Caption:=S+' '+IniFileFullPath;
  DisplayCaptionOfStringGrid;
  {$IFDEF Japanese}
  Jap:=TPzConv.Create(Nil);
  {$ENDIF Japanese}
  Field1List:=TStringList.Create;
  Field2List:=TStringList.Create;
  Field3List:=TStringList.Create;
  StringGrid.RowCount:=2;
  EnableOrDisableControls(rsIdle);
end;



procedure TFormMain.FormDestroy(Sender: TObject);
var
  S: String;
begin
  if Connected=True then
  begin
    IdPOP3.Disconnect;
    S:='Program disconnected the connection automaticaly.';
    Logging(S);
  end;
  WriteParametersToIniFile(IniFileFullPath);
  FormStatus.WriteFormStatusToIniFile(FormMain, IniFileFullPath);
  WriteStringGridStatusToIniFile(IniFileFullPath);
  WriteFontInfoToIniFile(FormMain, IniFileFullPath);
  {$IFDEF Japanese}
  Jap.Free;
  {$ENDIF Japanese}
  Field1List.Free;
  Field2List.Free;
  Field3List.Free;
  S:='Program finished.';
  Logging(S);
end;



procedure TFormMain.MainMenuOptionSettingsClick(Sender: TObject);
begin
  FormSettings.Font:=FormMain.Font;
  FormSettings.Show;
end;

procedure TFormMain.MainMenuOptionFontClick(Sender: TObject);
var
  I: Integer;
begin
  FontDialog.Font:=Font;
  if FontDialog.Execute=True then
  begin
    Font:= FontDialog.Font;
    for I:= 0 to Screen.FormCount-1 do
    begin
      Screen.Forms[I].Font:=Font;
      //Screen.Forms[I].Dispatch(Msg);
    end;
    //FontInfo.WriteFontInfoToIniFile(FormMain, IniFileFullPath);
  end;
end;



procedure SplitStringListToHeaderAndBody(MailTop: TStringList; MailHeader: TIdMessage; MailBody: TStringList);
var
  I: Integer;
  S: String;
begin
  I:=0;
  repeat
    S:=MailTop[I];
    if S<>'' then
    begin
      MailHeader.Headers.Add(S);
    end;
    inc(I);
  until S='';
  while I<=MailTop.Count-1 do
  begin
    S:=MailTop[I];
    MailBody.Add(S);
    inc(I)
  end;
end;



procedure ConvertMailBodyFromJISToSJIS(MailBody: TStringList);
var
  I:               Integer;
begin
  for I:=0 to MailBody.Count-1 do
  begin
    MailBody[I]:=Jap.JisToSJis(MailBody[I]);
  end;
end;



procedure TFormMain.StringGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var
  LoadFileName: String;
begin
  if (RunningStatus<>rsOnLineBrowsing)and(RunningStatus<>rsOffLineBrowsing) then
    Exit;
  if TabControlMain.TabIndex=0 then
    LoadFileName:=LogDir+'\'+StringGrid.Cells[2, ARow]+'-B.txt'
  else
    LoadFileName:=LogDir+'\'+StringGrid.Cells[2, ARow]+'-H.txt';
  try
    Memo.Lines.LoadFromFile(LoadFileName);
  except
    Memo.Lines.Clear;
    Memo.Lines.Add('Could not load file '+LoadFileName);
    Memo.Lines.Add('Error position is StringGridSelectCell');
    Memo.Lines.Add('Row='+IntToStr(ARow));
  end;
  EditActionReason.Text:=StringGrid.Cells[1, ARow];
end;



procedure TFormMain.TabControlMainChange(Sender: TObject);
var
  LoadFileName: String;
  ARow: Integer;
begin
  ARow:=StringGrid.Selection.Top;
  if TabControlMain.TabIndex=0 then
    LoadFileName:=LogDir+'\'+StringGrid.Cells[2, ARow]+'-B.txt'
  else
    LoadFileName:=LogDir+'\'+StringGrid.Cells[2, ARow]+'-H.txt';
  try
    Memo.Lines.LoadFromFile(LoadFileName);
  except
    Memo.Lines.Clear;
    Memo.Lines.Add('Could not load file '+LoadFileName);
    Memo.Lines.Add('Error position is TabControlMainChange');
    Memo.Lines.Add('Row='+IntToStr(ARow));
  end;
end;



procedure TFormMain.TimerNoopTimer(Sender: TObject);
begin
  TimerNoop.Enabled:=False;
  if Connected=True then
  begin
    IdPOP3.KeepAlive;
  end;
  TimerNoop.Enabled:=True;
  TimeredDisplayInStatusBar('Sent NOOP command.', 3);
end;



procedure TFormMain.MainMenuFileAbortClick(Sender: TObject);
begin
  ForcedTermination:=True;
end;



procedure TFormMain.MainMenuFileDisconnectClick(Sender: TObject);
var
  S: String;
begin
  IdPOP3.Disconnect;
  Connected:=False;
  TimerNoop.Enabled:=False;
  DisplayConnectionStatus(Connected);
  RunningStatus:=rsOffLineBrowsing;
  EnableOrDisableControls(RunningStatus);
  S:='Disconnected by user.';
  StatusBar.SimpleText:=S;
  Logging(S);
end;



procedure TFormMain.EnableOrDisableControls(RunningStatus: TMyRunningStatus);
begin
  if RunningStatus=rsIdle then
  begin
    MainMenuFile.Enabled:=True;
    MainMenuFileCheck.Enabled:=True;
    MainMenuFileDelete.Enabled:=False;
    MainMenuFileDeleteTest.Enabled:=False;
    MainMenuFileAbort.Enabled:=False;
    MainMenuFileDisconnect.Enabled:=False;
    MainMenuFileExit.Enabled:=True;
    MainMenuEdit.Enabled:=True;
    MainMenuSearch.Enabled:=False;
    MainMenuSort.Enabled:=False;
    MainMenuOption.Enabled:=True;
  end
  else if RunningStatus=rsChecking then
  begin
    MainMenuFile.Enabled:=True;
    MainMenuFileCheck.Enabled:=False;
    MainMenuFileDelete.Enabled:=False;
    MainMenuFileDeleteTest.Enabled:=False;
    MainMenuFileAbort.Enabled:=True;
    MainMenuFileDisconnect.Enabled:=False;
    MainMenuFileExit.Enabled:=False;
    MainMenuEdit.Enabled:=False;
    MainMenuSearch.Enabled:=False;
    MainMenuSort.Enabled:=False;
    MainMenuOption.Enabled:=False;
  end
  else if RunningStatus=rsOnLineBrowsing then
  begin
    MainMenuFile.Enabled:=True;
    MainMenuFileCheck.Enabled:=False;
    MainMenuFileDelete.Enabled:=True;
    MainMenuFileDeleteTest.Enabled:=True;
    MainMenuFileAbort.Enabled:=False;
    MainMenuFileDisconnect.Enabled:=True;
    MainMenuFileExit.Enabled:=True;
    MainMenuEdit.Enabled:=True;
    MainMenuSearch.Enabled:=True;
    MainMenuSort.Enabled:=True;
    MainMenuOption.Enabled:=True;
  end
  else if RunningStatus=rsOffLineBrowsing then
  begin
    MainMenuFile.Enabled:=True;
    MainMenuFileCheck.Enabled:=True;
    MainMenuFileDelete.Enabled:=False;
    MainMenuFileDeleteTest.Enabled:=False;
    MainMenuFileAbort.Enabled:=False;
    MainMenuFileDisconnect.Enabled:=False;
    MainMenuFileExit.Enabled:=True;
    MainMenuEdit.Enabled:=True;
    MainMenuSearch.Enabled:=True;
    MainMenuSort.Enabled:=True;
    MainMenuOption.Enabled:=True;
  end
  else if RunningStatus=rsDeleting then
  begin
    MainMenuFile.Enabled:=True;
    MainMenuFileCheck.Enabled:=False;
    MainMenuFileDelete.Enabled:=False;
    MainMenuFileDeleteTest.Enabled:=False;
    MainMenuFileAbort.Enabled:=True;
    MainMenuFileDisconnect.Enabled:=False;
    MainMenuFileExit.Enabled:=False;
    MainMenuEdit.Enabled:=False;
    MainMenuSearch.Enabled:=False;
    MainMenuSort.Enabled:=False;
    MainMenuOption.Enabled:=False;
  end
  else if RunningStatus=rsSorting then
  begin
    MainMenuFile.Enabled:=False;
    MainMenuEdit.Enabled:=False;
    MainMenuSearch.Enabled:=False;
    MainMenuSort.Enabled:=False;
    MainMenuOption.Enabled:=False;
  end
end;



procedure TFormMain.TimerClearStatusTimer(Sender: TObject);
begin
  TimerClearStatus.Enabled:=False;
  StatusBar.SimpleText:='';
end;



procedure TFormMain.MainMenuFileDeleteClick(Sender: TObject);
begin
  DeleteMails(False); (* Delete mails actually becase ShamMode is False*)
end;



procedure TFormMain.MainMenuFileDeleteTestClick(Sender: TObject);
begin
  DeleteMails(True); (* Do not delete mails actually because ShamMode is True *)
end;



procedure TFormMain.SelectRow(Row: Integer);
var
  LoadFileName: String;
begin
    StringGrid.Selection:=TGridRect(Rect(0, Row, StringGrid.ColCount-1, Row));
    EditActionReason.Text:=StringGrid.Cells[1, Row];
    if TabControlMain.TabIndex=0 then
    begin
      LoadFileName:=LogDir+'\'+StringGrid.Cells[2, Row]+'-B.txt'
    end
    else
    begin
      LoadFileName:=LogDir+'\'+StringGrid.Cells[2, Row]+'-H.txt';
    end;
    try
      Memo.Lines.LoadFromFile(LoadFileName);
    except
      Memo.Lines.Clear;
      Memo.Lines.Add('Could not load file '+LoadFileName);
      Memo.Lines.Add('Error position is SelectRow');
      Memo.Lines.Add('Row='+IntToStr(Row));
    end;
end;



procedure TFormMain.MainMenuEditRuleFileClick(Sender: TObject);
begin
  ShellExecute(Self.Handle, Pchar('Open'), PChar(RuleFileName), PChar(0), nil, SW_NORMAL);
end;



procedure TFormMain.MainMenuSearchTopClick(Sender: TObject);
begin
  StringGrid.TopRow:=1;
  SelectRow(1);
end;



procedure TFormMain.MainManuSearchLastClick(Sender: TObject);
var
  Row: Integer;
begin
  Row:=StringGrid.RowCount-1;
  StringGrid.TopRow:=Row;
  SelectRow(Row);
end;




procedure TFormMain.DeleteMails(ShamMode: Boolean);
var
  Action: Integer;
  BGColor: String;
  BodyFileName: String;
  HTMLFile: TextFile;
  HeaderFileName: String;
  HTMLFileName: String;
  I: Integer;
  NumberOfMails: Integer;
  S: String;
begin
  S:=FormatDateTime('yyyymmdd-hhnnss', Now);
  HTMLFileName:=LogDir+'\'+S+'-DEL.html';
  LogFileName:=LogDir+'\'+S+'-DEL.txt';

  AssignFile(HTMLFile, HTMLFileName);
  try
    Rewrite(HTMLFile);
  except
    S:='Could not create the HTML file ('+HTMLFileName+').';
    Logging(S);
    StatusBar.SimpleText:=S;
    MessageDlg(S, mtWarning, [mbOK], 0);
    Exit;
  end;
  TimerNoop.Enabled:=False;
  RunningStatus:=rsDeleting;
  EnableOrDisableControls(RunningStatus);
  NumberOfMails:=StringGrid.RowCount-1;
  if NumberOfMails=0 then
    Exit;
  if ShamMode=True then
    S:='DeleteMode: Sham mode'
  else
    S:='DeleteMode: Exactly delete mode';
  Logging(S);
  ForcedTermination:=False; (* ForcedTermination is a global variable in this form *)
  S:='Number of mails is '+IntToStr(NumberOfMails)+'.';
  Logging(S);
  S:='MailUtl status log  '+HTMLFileName;
  Writeln(HTMLFile, '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">');
  Writeln(HTMLFile, '<HTML>');
  Writeln(HTMLFile, '<HEAD>');
  Writeln(HTMLFile, '<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=Shift_JIS">');
  Writeln(HTMLFile, '<TITLE>'+S+'</TITLE>');
  Writeln(HTMLFile, '</HEAD>');
  Writeln(HTMLFile, '<BODY>');
  Writeln(HTMLFile, '<H1>'+EncodeForHTML(S)+'</H1>');
  Writeln(HTMLFile, '<TABLE BORDER="1" CELLSPACING="0">');
  Writeln(HTMLFile, '<TR><TH>No.</TH><TH>Headers, part of body, and action</TH></TR>');
  I:=1;
  ProgressBar.Min:=1;
  ProgressBar.Max:=StringGrid.RowCount;
  while (I<=NumberOfMails)and(ForcedTermination=False) do
  begin (* Mail deleting loop *)
    StatusBar.SimpleText:='Processing '+StringGrid.Cells[4, I];
    ProgressBar.Position:=I;
    StringGrid.TopRow:=I;
    SelectRow(I);
    HeaderFileName:=StringGrid.Cells[2, I]+'-H.txt';
    BodyFileName:=StringGrid.Cells[2, I]+'-B.txt';
    Action:=StrToIntDef(StringGrid.Cells[0, I], ord(actNop));
    case Action of
      ord(actNop):           BGColor:='#FFFFFF';
      ord(actKeep):          BGColor:='#E0FFE0';
      ord(actDel):           BGColor:='#FFD0D0';
    end;
    S:=StringGrid.Cells[4, I]+'/'+IntToStr(NumberOfMails);
    Logging(S);
    Writeln(HTMLFile, '<TR><TD BGCOLOR="'+BGColor+'">'+S+'</TD><TD BGCOLOR="'+BGColor+'">');
    Writeln(HTMLFile, '<P>');
    S:='Subject: '+StringGrid.Cells[6, I];
    Logging(S);
    Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
    S:='From: '+StringGrid.Cells[7, I];
    Logging(S);
    Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
    case Action of
      ord(actNop):           S:='Action: Nop.';
      ord(actKeep):          S:='Action: Keep';
      ord(actDel):           S:='Action: Del';
    end;
    if Action>ord(ActNop) then
    begin
      S:=S+'  Reason: '+StringGrid.Cells[1, I];
    end;
    Writeln(HTMLFile, '<P>'+EncodeForHTML(S)+'</P>');
    Writeln(HTMLFile, '<P><A HREF="'+HeaderFileName+'">Full Headers</A> and <A HREF="'+BodyFileName+'">Part of Body</A></P>');
    if Action>ord(ActKeep) then
    begin
      if ShamMode=False then
      begin
        try
          IdPOP3.Delete(StrToInt(StringGrid.Cells[4, I]));
        except
          Logging('*** Error: Could not delete the mail ***');
        end;
      end;
    end;
    if ForcedTermination=False then
    begin
      inc(I);
    end;
    IdPOP3.KeepAlive;
    Application.ProcessMessages;
  end; (* Mail deleting loop *)
  Writeln(HTMLFile, '</TABLE>');
  Writeln(HTMLFile, '<P>End of MailUtl status log.</P>');
  Writeln(HTMLFile, '</BODY>');
  Writeln(HTMLFile, '</HTML>');
  CloseFile(HTMLFile);
  ProgressBar.Position:=0;
  TimerNoop.Enabled:=False;
  IdPOP3.Disconnect;
  Connected:=False;
  DisplayConnectionStatus(Connected);
  RunningStatus:=rsOffLineBrowsing;
  EnableOrDisableControls(RunningStatus);
  if ForcedTermination=True then
  begin
    S:='Deleting mails are forced terminated.';
  end
  else begin
    S:='Deleting mails are normaly finished.';
    StringGrid.RowCount:=2;
    EditActionReason.Text:='';
    for I:=0 to StringGrid.ColCount-1 do
    begin
      StringGrid.Cells[I, 1]:='';
    end;
    Memo.Lines.Clear;
  end;
  TimeredDisplayInStatusBar(S, 3);
  Logging(S);
end;


procedure TFormMain.CheckMailBox;
var
  Action_: Integer;
  CheckResult: TMyEvaluated;
  EscapedMsgId: String;
  I: Integer;
  MailBody: TStringList;
  MailHeader: TIdMessage;
  MailTop: TStringList;
  MsgSize: Integer;
  NumberOfMessages: Integer; (* Number of messages (e-mails) in the POP server *)
  S: String;
  SaveFileName: String;
  TopRow: Integer;
begin
  S:=FormatDateTime('yyyymmdd-hhnnss', Now);
  LogFileName:=Logdir+'\'+S+'-CHK.txt';
  (* Exit this procedure if password is empty. *)
  if POP3_Password='' then
  begin
    MessageDlg('Password is empty.', mtWarning, [mbOK], 0);
    Exit;
  end;

  EnableOrDisableControls(rsChecking);

  (* Start of this routine *)
  S:='Checking is started.';
  Logging(S);
  IdPOP3.Host:=POP3_Host;
  IdPOP3.Port:=POP3_Port;
  IdPOP3.Username:=POP3_Username;
  IdPOP3.Password:=POP3_Password;

  (* Connect to the POP server *)
  S:='Connecting to the host: '+POP3_Host+'  Port: '+IntToStr(POP3_Port)+'...';
  StatusBar.SimpleText:=S;
  Logging(S);
  try
    IdPOP3.Connect;
  except
    on E: IdReplyPOP3.EIdReplyPOP3Error do begin
      S:='Error code: '+E.ErrorCode +' Enhanced code: '+E.EnhancedCode;
      MessageDlg(S, mtWarning, [mbOK], 0);
      Logging(S);
      EnableOrDisableControls(rsIdle);
      Exit;
    end;
    on E: EIdExceptionBase do begin
      S:='Already connected.';
      MessageDlg(S, mtWarning, [mbOK], 0);
      Logging(S);
      EnableOrDisableControls(rsOnLineBrowsing);
      Exit;
    end;
  end;

  (* Obtain the number of messages (e-mails) in the POP server *)
  (* Exit this procedure if it is zero. *)
  NumberOfMessages:=IdPOP3.CheckMessages;
  if NumberOfMessages=0 then
  begin
    S:='No message in the POP server';
    MessageDlg(S, mtInformation, [mbOK], 0);
    Logging(S);
    IdPOP3.Disconnect;
    EnableOrDisableControls(rsIdle);
    Exit;
  end;

  Connected:=True;

  TimerNoop.Enabled:=False;
  DisplayConnectionStatus(Connected);
  S:='Number of messages is '+IntToStr(NumberOfMessages);
  Logging(S);

  (* Obtain the rules from the rulefile named "RuleFileName" *)
  if ReadRuleFromFile(RuleFileName)=False then
  begin
    S:='Could not read from rules from '+RuleFileName+'.';
    MessageDlg(S, mtWarning, [mbOK], 0);
    Logging(S);
    EnableOrDisableControls(rsIdle);
    Exit;
  end;
  S:='MailUtl rules are loaded.';
  Logging(S);

  (* Create Mail message string list *)
  MailHeader:=TIdMessage.Create;
  MailBody:=TStringList.Create;
  MailTop:=TStringList.Create;

  StringGrid.RowCount:=2;
  I:=1;
  ProgressBar.Min:=1;
  ProgressBar.Max:=NumberOfMessages;
  ForcedTermination:=False;
  while (I<=NumberOfMessages)and(ForcedTermination=False) do
  begin (* Mail retrieving loop *)
    MailHeader.Clear;
    MailBody.Clear;
    MailTop.Clear;
    Memo.Clear;
    CheckResult.Action:=actNop;
    CheckResult.Reason:='';
    Action_:=-1;
    ProgressBar.Position:=I;
    S:='No. '+IntToStr(I)+'/'+IntToStr(NumberOfMessages);
    Logging(S);
    StatusBar.SimpleText:=S;
    S:='Getting message size';
    StatusBar.SimpleText:=S;
    MsgSize:=IdPOP3.RetrieveMsgSize(I);
    S:='Message size is '+IntToStr(MsgSize);
    StatusBar.SimpleText:=S;
    S:='Sending TOP command.';
    StatusBar.SimpleText:=S;
    try
      IdPOP3.Top(I, MailTop, MaxLines);
    except
      S:='The TOP command is failed.  The program will abort.';
      StatusBar.SimpleText:=S;
      Logging(S);
      ForcedTermination:=True;
    end;
    if ForcedTermination=False then
    begin
      SplitStringListToHeaderAndBody(MailTop, MailHeader, MailBody);
      ArrangeHeaders(MailHeader);
      {$IFDEF Japanese}
      ConvertMailBodyFromJISToSJIS(MailBody);
      {$ENDIF Japanese}

      (* MailUtl original rule check *)
      if UseMailUtlRules=True then
      begin
        CheckResult:=CheckRule(MailHeader, MsgSize);
        Action_:=Ord(CheckResult.Action);
        EditActionReason.Text:=CheckResult.Reason;
      end;

      if Action_>=0 then
         StringGrid.Cells[0, I]:=IntToStr(Action_);

      if I>1 then
      begin
        StringGrid.RowCount:=I+1;
        StringGrid.Selection:=TGridRect(Rect(0, I, StringGrid.ColCount-1, I));
      end;
      StringGrid.Cells[1, I]:=CheckResult.Reason;
      EscapedMsgId:=GetFieldValueInList(MailHeader.Headers, 'Message-ID:');

      (* Remove letters what can not use in a file name *)
      brx.Subst('s/[\\\/:,;*?"<>|]//g', EscapedMsgId);

      StringGrid.Cells[2, I]:=EscapedMsgId;
      StringGrid.Cells[4, I]:=IntToStr(I);
      if Action_>=0 then
      begin
        case Action_ of
          ord(actNop):  StringGrid.Cells[5, I]:='';
          ord(actKeep): StringGrid.Cells[5, I]:='Keep';
          ord(actDel):  StringGrid.Cells[5, I]:='Delete';
        end
      end;
      StringGrid.Cells[6, I]:=MailHeader.Subject;
      StringGrid.Cells[7, I]:=MailHeader.From.Text;
      StringGrid.Cells[8, I]:=FormatDateTime('yyyy/mm/dd hh:nn:ss', MailHeader.Date);
      StringGrid.Cells[9, I]:=FormatFloat('#,##0', MsgSize);

      if I>StringGrid.VisibleRowCount then
      begin
        TopRow:=I-(StringGrid.VisibleRowCount-1);
        if TopRow>1 then
        begin
          StringGrid.TopRow:=TopRow;
        end;
      end;

      if TabControlMain.TabIndex=0 then
      begin
        Memo.Lines:=MailBody;
      end
      else begin
        Memo.Lines:=MailHeader.Headers;
      end;

      (* Save header to file *)
      try
        SaveFileName:=LogDir+'\'+EscapedMsgId+'-H.txt';
        S:='Saving headers into the file '+SaveFileName+'...';
        StatusBar.SimpleText:=S;
        MailHeader.Headers.SaveToFile(SaveFileName);
        //SaveStringListToFile(MailHeader.Headers, SaveFileName);
      except
        Logging('*** Error: Could not save file '+SaveFileName+' ***');
      end;

      (* Save part of body to file *)
      try
        SaveFileName:=LogDir+'\'+EscapedMsgId+'-B.txt';
        S:='Saving headers into the file '+SaveFileName+'...';
        StatusBar.SimpleText:=S;
        MailBody.SaveToFile(SaveFileName);
        //SaveStringListToFile(MailBody, SaveFileName);
      except
        Logging('*** Error: Could not save file '+SaveFileName+' ***');
      end;
    end;
    inc(I);
    Application.ProcessMessages;
  end; (* Mail retrieving loop *)

  (* At this point, KeepAlive timer is still disabled *)

  (* Release Mail message string list *)
  MailHeader.Free;
  MailBody.Free;
  MailTop.Free;
  ProgressBar.Position:=0;
  Memo.Lines.Clear;

  if StringGrid.RowCount>1 then
  begin
    StringGrid.TopRow:=1;
    SelectRow(1);
  end;

  (* At this point, KeepAlive timer is still disabled *)

  if ForcedTermination=True then
  begin
    IdPOP3.Disconnect;
    Connected:=False;
    DisplayConnectionStatus(Connected);
    RunningStatus:=rsOffLineBrowsing;
    EnableOrDisableControls(RunningStatus);
    S:='Checking is forced terminated.';
    Logging(S);
  end
  else begin
    TimerNoop.Enabled:=True;
    RunningStatus:=rsOnLineBrowsing;
    EnableOrDisableControls(rsOnLineBrowsing);
    S:='Checking is normaly finished. '+IntToStr(NumberOfMessages)+' mails.';
    Logging(S);
    StatusBar.SimpleText:=S;
  end;
end;


procedure TFormMain.MainMenuSortByActionClick(Sender: TObject);
var
  LastRunningStatus: TMyRunningStatus;
  Row1: Integer;
  Row2: Integer;
  TmpList: TStringList;
begin
  LastRunningStatus:=RunningStatus;
  RunningStatus:=rsSorting;
  EnableOrDisableControls(RunningStatus);
  if StringGrid.RowCount<=2 then
    Exit;
  EditActionReason.Text:='';
  Memo.Clear;
  StringGrid.Visible:=False;
  TmpList:=TStringList.Create;
  ProgressBar.Min:=1;
  ProgressBar.Max:=StringGrid.RowCount-1;
  for Row1:=1 to StringGrid.RowCount-2 do
    begin
      ProgressBar.Position:=Row1;
      for Row2:=Row1+1 to StringGrid.RowCount-1 do
      begin
        if StrToIntDef(StringGrid.Cells[0, Row1], 0) < StrToIntDef(StringGrid.Cells[0, Row2], 0) then
        begin
          TmpList.Assign(StringGrid.Rows[Row1]);
          StringGrid.Rows[Row1]:=StringGrid.Rows[Row2];
          StringGrid.Rows[Row2]:=TmpList;
        end;
        Application.ProcessMessages;
      end;
    end;
  TmpList.Free;
  StringGrid.Visible:=True;
  ProgressBar.Position:=0;
  StringGrid.TopRow:=1;
  SelectRow(1);
  TimeredDisplayInStatusBar('Finished sorting by Action.', 3);
  RunningStatus:=LastRunningStatus;
  EnableOrDisableControls(RunningStatus);
end;

procedure TFormMain.MainMenuSortByNumberClick(Sender: TObject);
var
  LastRunningStatus: TMyRunningStatus;
  Row1: Integer;
  Row2: Integer;
  TmpList: TStringList;
begin
  LastRunningStatus:=RunningStatus;
  RunningStatus:=rsSorting;
  EnableOrDisableControls(RunningStatus);
  if StringGrid.RowCount<=2 then
    Exit;
  EditActionReason.Text:='';
  Memo.Clear;
  StringGrid.Visible:=False;
  TmpList:=TStringList.Create;
  ProgressBar.Min:=1;
  ProgressBar.Max:=StringGrid.RowCount-1;
  for Row1:=1 to StringGrid.RowCount-2 do
    begin
      for Row2:=Row1+1 to StringGrid.RowCount-1 do
      begin
        ProgressBar.Position:=Row1;
        if StrToIntDef(StringGrid.Cells[4, Row1], 0) > StrToIntDef(StringGrid.Cells[4, Row2], 0) then
        begin
          TmpList.Assign(StringGrid.Rows[Row1]);
          StringGrid.Rows[Row1]:=StringGrid.Rows[Row2];
          StringGrid.Rows[Row2]:=TmpList;
        end;
        Application.ProcessMessages;
      end;
    end;
  TmpList.Free;
  StringGrid.Visible:=True;
  ProgressBar.Position:=0;
  StringGrid.TopRow:=1;
  SelectRow(1);
  TimeredDisplayInStatusBar('Finished sorting by Number.', 3);
  RunningStatus:=LastRunningStatus;
  EnableOrDisableControls(RunningStatus);
end;

procedure TFormMain.MainMenuFileCheckClick(Sender: TObject);
begin
  CheckMailBox;
end;

end.
